@@ -50,9 +50,6 @@ static void epoll_event_arm(int fd, int events)
 static void epoll_event_delete(int fd)
 {
 	DEBUG(2, "epoll_event_delete(fd=%d)", fd);
-#if 0
-	epoll_ctl(efd, EPOLL_CTL_DEL, fd, NULL);
-#endif
 }
 
 static void epoll_event_wait(void)
@@ -75,6 +72,8 @@ static int epoll_event_fd(int *revents)
 	DEBUG(3, "\tepoll_ev[%d] = {revents=%d, data.fd=%d}", pindex, epoll_ev[pindex].events, epoll_ev[pindex].data.fd);
         if (epoll_ev[pindex].events & EPOLLIN) events |= EVENT_READ;
 	if (epoll_ev[pindex].events & EPOLLOUT) events |= EVENT_WRITE;
+	if (epoll_ev[pindex].events & EPOLLERR) events |= EVENT_ERR;
+	if (events == 0) DEBUG(2, "events for fd %d = %d", epoll_ev[pindex].data.fd, epoll_ev[pindex].events);
 	*revents = events;
 	return epoll_ev[pindex].data.fd;
 }
