@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2000-2015  Ulric Eriksson <ulric@siag.nu>
+   Copyright (C) 2000-2016  Ulric Eriksson <ulric@siag.nu>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -107,7 +107,7 @@ static int port;
 
 static int control_acl;
 static char *ctrlport = NULL;
-int listenfd;
+int listenfd = -1;
 static int ctrlfd = -1;
 static char *jail = NULL;
 static char *user = NULL;
@@ -531,25 +531,37 @@ static int rewrite_request(int i, int n, char *b)
 		q = strstr(b, "\n\n");
 	}
 	if (!q) return n;		/* not a header */
-#if 0	/* how is that supposed to happen? */
-	if (q >= b+n) return n;		/* outside of buffer */
-#endif
+
 	/* Look for existing X-Forwarded-For */
 	DEBUG(2, "Looking for X-Forwarded-For");
-
-	if (pen_strcasestr(b, "\nX-Forwarded-For:")) return n;
-
-	DEBUG(2, "Adding X-Forwarded-For");
-	/* Didn't find one, add our own */
-	snprintf(p, sizeof p, "\r\nX-Forwarded-For: %s",
-		pen_ntoa(&clients[conns[i].client].addr));
-	pl=strlen(p);
-	if (n+pl > BUFFER_MAX) return n;
-
-	memmove(q+pl, q, b+n-q);
-	memmove(q, p, pl);
-
-	n += pl;
+	if (!pen_strcasestr(b, "\nX-Forwarded-For:"))
+	{
+		DEBUG(2, "Adding X-Forwarded-For");
+		/* Didn't find one, add our own */
+		snprintf(p, sizeof p, "\r\nX-Forwarded-For: %s",
+			pen_ntoa(&clients[conns[i].client].addr));
+		pl=strlen(p);
+		if (n+pl > BUFFER_MAX) return n;
+		memmove(q+pl, q, b+n-q);
+		memmove(q, p, pl);
+		n += pl;
+	}
+	b[n] = '\0';
+	if (!pen_strcasestr(b, "\nX-Forwarded-Proto:")){
+		DEBUG(2, "Adding X-Forwarded-Proto");
+		/* Didn't find one, add our own */
+		#ifdef HAVE_LIBSSL
+		snprintf(p, sizeof p, "\r\nX-Forwarded-Proto: %s",
+			(conns[i].ssl)?"https":"http");
+		#else
+		snprintf(p, sizeof p, "\r\nX-Forwarded-Proto: %s","http");
+		#endif  /* HAVE_LIBSSL */
+		pl=strlen(p);
+		if (n+pl > BUFFER_MAX) return n;
+		memmove(q+pl, q, b+n-q);
+		memmove(q, p, pl);
+		n += pl;
+	}
 	return n;
 }
 
@@ -697,6 +709,7 @@ static int copy_up(int i)
 		}
 
 		n = my_send(to, b, rc, 0);	/* no ssl here */
+		conns[i].state &= ~CS_HALFDEAD;
 		SAVE_ERRNO;
 
 		DEBUG(2, "copy_up: send(%d, %p, %d, 0) returns %d, socket_errno = %d",
@@ -775,11 +788,8 @@ static int copy_down(int i)
 		int n;
 
 		if (udp) {
-			struct sockaddr_storage *ss = &clients[conns[i].client].addr;
-			socklen_t sss = pen_ss_size(ss);
-			DEBUG(2, "copy_down sending %d bytes to socket %d", rc, to);
-			n = sendto(to, (void *)b, rc, 0, (struct sockaddr *)ss, sss);
-			close_conn(i);
+			n = send(to, (void *)b, rc, 0);
+			conns[i].state = CS_CONNECTED;
 			return 0;
 		}
 
@@ -922,17 +932,12 @@ static void init(int argc, char **argv)
 
 	DEBUG(2, "init(%d, %p); port = %d", argc, argv, port);
 
-#if 0
-	conns = pen_calloc(connections_max, sizeof *conns);
-	clients = pen_calloc(clients_max, sizeof *clients);
-#else
 	debug("Before: conns = %p, connections_max = %d, clients = %p, clients_max = %d",
 		conns, connections_max, clients, clients_max);
 	if (connections_max == 0) expand_conntable(CONNECTIONS_MAX);
 	if (clients_max == 0) expand_clienttable(CLIENTS_MAX);
 	debug("After: conns = %p, connections_max = %d, clients = %p, clients_max = %d",
 		conns, connections_max, clients, clients_max);
-#endif
 
 	current = 0;
 
@@ -981,14 +986,6 @@ static void init(int argc, char **argv)
 		clients[i].csx = 0;
 		clients[i].crx = 0;
 	}
-#if 0
-	for (i = 0; i < connections_max; i++) {
-		conns[i].upfd = -1;
-		conns[i].downfd = -1;
-		conns[i].upn = 0;
-		conns[i].downn = 0;
-	}
-#endif
 
 	if (debuglevel) {
 		debug("%s starting", PACKAGE_STRING);
@@ -1103,6 +1100,9 @@ static int open_listener(char *a)
 
 	setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (void *)&one, sizeof one);
 	setsockopt(listenfd, SOL_SOCKET, SO_KEEPALIVE, (void *)&optval, sizeof optval);
+#ifdef SO_REUSEPORT
+	setsockopt(listenfd, SOL_SOCKET, SO_REUSEPORT, (void *)&one, sizeof one);
+#endif
 
 	if (bind(listenfd, (struct sockaddr *)&ss, pen_ss_size(&ss)) < 0) {
 		error("can't bind local address");
@@ -1237,12 +1237,6 @@ static void do_cmd(char *b, void (*output)(void *, char *, ...), void *op)
 				} else {
 					ma = "128";
 				}
-#if 0
-				if (inet_pton(AF_INET6, ip, ipaddr) != 1) {
-					debug("acl: can't convert address %s", ip);
-					return;
-				}
-#else
 				struct sockaddr_storage ss;
 				struct sockaddr_in6 *si6;
 				if (pen_aton(ip, &ss) != 1) {
@@ -1255,7 +1249,6 @@ static void do_cmd(char *b, void (*output)(void *, char *, ...), void *op)
 				}
 				si6 = (struct sockaddr_in6 *)&ss;
 				memcpy(ipaddr, &si6->sin6_addr, sizeof ipaddr);
-#endif
 				add_acl_ipv6(a, ipaddr, atoi(ma), permit);
 			} else {
 				struct in_addr ipaddr, mask;
@@ -1723,7 +1716,10 @@ static void add_client(int downfd, struct sockaddr_storage *cli_addr)
 
 	/* we don't know the client address for udp until we read the message */
 	if (udp) {
+		int n, one = 1;
 		socklen_t len = sizeof *cli_addr;
+		struct sockaddr listenaddr;
+		socklen_t listenlen = sizeof listenaddr;
 		rc = recvfrom(listenfd, (void *)b, sizeof b, 0, (struct sockaddr *)cli_addr, &len);
 		DEBUG(2, "add_client: received %d bytes from client", rc);
 		if (rc < 0) {
@@ -1731,7 +1727,35 @@ static void add_client(int downfd, struct sockaddr_storage *cli_addr)
 				debug("Error receiving data");
 			return;
 		}
+	/* we need a downfd for udp as well */
+		downfd = socket_nb(cli_addr->ss_family, protoid, 0);
+		if (downfd == -1) {
+			debug("Can't create downfd");
+			return;
+		}
+#ifdef SO_REUSEPORT
+		setsockopt(downfd, SOL_SOCKET, SO_REUSEPORT, &one, sizeof one);
+#endif
+		n = getsockname(listenfd, &listenaddr, &listenlen);
+		if (n != 0) {
+			debug("getsockname returns %d, errno = %d", n, errno);
+			close(downfd);
+			return;
+		}
+		n = bind(downfd, &listenaddr, listenlen);
+		if (n != 0) {
+			debug("bind returns %d, errno = %d", n, errno);
+			close(downfd);
+			return;
+		}
+		n = connect(downfd, (struct sockaddr *)cli_addr, pen_ss_size(cli_addr));
+		if (n != 0) {
+			debug("connect (downfd = %d) returns %d, errno = %d", downfd, n, errno);
+			close(downfd);
+			return;
+		}
 	}
+
 	client = store_client(cli_addr);	// no server yet
 	DEBUG(2, "store_client returns %d", client);
 
@@ -1772,6 +1796,7 @@ static void add_client(int downfd, struct sockaddr_storage *cli_addr)
 	if (udp && rc > 0) {	/* pass on the message */
 		/* we are "connected" and don't need sendto */
 		rc = send(conns[conn].upfd, (void *)b, rc, 0);
+		conns[conn].state &= ~CS_HALFDEAD;
 		DEBUG(2, "add_client: wrote %d bytes to socket %d", rc, conns[conn].upfd);
 	}
 }
@@ -1804,11 +1829,13 @@ static int flush_down(int i)
 	struct sockaddr name;
 	size_t size = sizeof(struct sockaddr);
 
-	if (!udp)
+	if (!udp) {
 		n = my_send(conns[i].downfd, conns[i].downbptr, conns[i].downn, 0);
-	else
+	} else {
 		n = sendto(conns[i].downfd, (void *)conns[i].downbptr, conns[i].downn, 0,
 			(struct sockaddr *) &name, size);
+		conns[i].state &= ~CS_HALFDEAD;
+	}
 	err = socket_errno;
 #endif  /* HAVE_LIBSSL */
 
@@ -1847,11 +1874,13 @@ static int flush_up(int i)
        struct sockaddr name;
        size_t size = sizeof(struct sockaddr);
 
-	if (!udp)
+	if (!udp) {
 		n = my_send(conns[i].upfd, conns[i].upbptr, conns[i].upn, 0);
-	else
+	} else {
 		n = sendto(conns[i].upfd, (void *)conns[i].upbptr, conns[i].upn, 0,
 			(struct sockaddr *) &name, size);
+		conns[i].state &= ~CS_HALFDEAD;
+	}
 	err = socket_errno;
 
 	DEBUG(2, "flush_up(%d): send(%d, %p, %d, 0) returns %d, errno = %d, socket_errno = %d", \
@@ -1876,20 +1905,6 @@ static int flush_up(int i)
 	return n;
 }
 
-/* For UDP, connection attempts to unreachable servers would sit in the
-   connection table forever unless we remove them by force.
-   This function is pretty brutal, it simply vacates the next slot
-   after the most recently used one. This will always succeed.
-*/
-static void recycle_connection(void)
-{
-	int i;
-
-	i = connections_last+1;
-	if (i >= connections_max) i = 0;
-	close_conn(i);
-}
-
 #ifndef WINDOWS
 static void setup_signals(void)
 {
@@ -1960,7 +1975,7 @@ static void check_listen_socket(void)
 		dsr_frame(listenfd);
 	} else if (udp) {
 		/* special case for udp */
-		downfd = listenfd;
+		downfd = 0;
 		add_client(downfd, &cli_addr);
 	} else {
 	/* process tcp connection(s) */
@@ -2100,7 +2115,9 @@ static void arm_listenfd(void)
 {
 	static int can_accept = 0;
 
-	if (can_accept) {
+	if (udp) {
+		event_arm(listenfd, EVENT_READ);
+	} else if (can_accept) {
 		if (connections_used >= connections_max) {
 			event_arm(listenfd, 0);
 			can_accept = 0;
@@ -2137,32 +2154,37 @@ static int handle_events(int *pending_close)
                 conn = fd2conn_get(fd);
 		DEBUG(3, "fd = %d => conn = %d", fd, conn);
 		if (conn == -1) continue;
-                if (conns[conn].state & CS_IN_PROGRESS) {
+
+		if (events & EVENT_ERR) {
+			DEBUG(2, "Error on fd %d, connection %d", fd, conn);
+			conns[conn].state = CS_CLOSED;
+		} else if (conns[conn].state & CS_IN_PROGRESS) {
                         if (fd == conns[conn].upfd && events & EVENT_WRITE) {
                                 check_if_connected(conn);
                         }
-                        continue;
-                }
-		conns[conn].t = now;
-                if (fd == conns[conn].downfd) {
-                        if (!udp && (events & EVENT_READ)) {
-                                if (!try_copy_up(conn)) closing = 1;
-                        }
-                        if (events & EVENT_WRITE) {
-                                if (!try_flush_down(conn)) closing = 1;
-                        }
-                } else {        /* down */
-                        if (events & EVENT_READ) {
-                                if (!try_copy_down(conn)) closing = 1;
-                        }
-                        if (!udp && (events & EVENT_WRITE)) {
-                                if (!try_flush_up(conn)) closing = 1;
-                        }
-                }
+                } else {
+			conns[conn].t = now;
+                	if (fd == conns[conn].downfd) {
+                        	if (events & EVENT_READ) {
+                                	if (!try_copy_up(conn)) closing = 1;
+                        	}
+                        	if (events & EVENT_WRITE) {
+                                	if (!try_flush_down(conn)) closing = 1;
+                        	}
+                	} else {        /* down */
+                        	if (events & EVENT_READ) {
+                                	if (!try_copy_down(conn)) closing = 1;
+                        	}
+                        	if (events & EVENT_WRITE) {
+                                	if (!try_flush_up(conn)) closing = 1;
+                        	}
+                	}
+		}
 		if (conns[conn].state == CS_CLOSED) {
 			DEBUG(2, "Connection %d was closed", conn);
 			closing = 1;
 		}
+
 		if (closing) {
 			pending_close[npc++] = conn;
 		}
@@ -2172,17 +2194,21 @@ static int handle_events(int *pending_close)
 
 static void pending_and_closing(int *pending_close, int npc)
 {
-	int j, p, start;
+	int j, p, start, npe;
 
 	if (pending_list != -1) {
+		npe = npc;
 		p = start = pending_list;
 		do {
 			int conn = dlist_value(p);
 			if (conns[conn].state == CS_IN_PROGRESS) {
-				check_if_timeout(conn);
+				pending_close[npe++] = conn;
 			}
 			p = dlist_next(p);
 		} while (p != start);
+		for (j = npc; j < npe; j++) {
+			check_if_timeout(pending_close[j]);
+		}
 	}
         for (j = 0; j < npc; j++) {
 		int conn = pending_close[j];
@@ -2233,7 +2259,6 @@ void mainloop(void)
         while (loopflag) {
                 check_signals();
 		if (dsr_if) dsr_arp(listenfd);
-                if (udp && (connections_used >= connections_max)) recycle_connection();
 		arm_listenfd();
                 event_wait();
                 now = time(NULL);
@@ -2277,7 +2302,8 @@ static int options(int argc, char **argv)
 			do_cmd(optarg, output_file, stdout);
 			break;
 		case 'Q':
-			event_init = kqueue_init;
+			fprintf(stderr, "Since Pen 0.26.0, kqueue is already the default on supported systems,\n"
+					"making the -Q option obsolete\n");
 			break;
 		case 'P':
 			event_init = poll_init;
@@ -2436,11 +2462,6 @@ int main(int argc, char **argv)
 	argc -= n;
 	argv += n;
 
-#ifndef WINDOWS
-	if (argc < 1) {
-		usage();
-	}
-#endif
 	now = time(NULL);
 #ifdef WINDOWS
 	start_winsock();
@@ -2449,6 +2470,10 @@ int main(int argc, char **argv)
 	read_cfg(cfgfile);
 
 #ifndef WINDOWS
+	if (listenfd == -1 && argc < 1) {
+		usage();
+	}
+
 	struct rlimit r;
 	getrlimit(RLIMIT_CORE, &r);
 	r.rlim_cur = r.rlim_max;
@@ -2478,15 +2503,17 @@ int main(int argc, char **argv)
 		proto = "udp";
 	}
 
-	snprintf(listenport, sizeof listenport, "%s", argv[0]);
-	/* Direct server return */
-	if (dsr_if) {
-		listenfd = dsr_init(dsr_if, listenport);
-		if (listenfd == -1) {
-			error("Can't initialize direct server return");
+	if (listenfd == -1) {
+		snprintf(listenport, sizeof listenport, "%s", argv[0]);
+		/* Direct server return */
+		if (dsr_if) {
+			listenfd = dsr_init(dsr_if, listenport);
+			if (listenfd == -1) {
+				error("Can't initialize direct server return");
+			}
+		} else {
+			listenfd = open_listener(listenport);
 		}
-	} else {
-		listenfd = open_listener(listenport);
 	}
 	init(argc, argv);
 
