@@ -375,7 +375,7 @@ int try_server(int index, int conn)
 			conns[conn].state |= CS_CLOSED_DOWN;
 		}
 		event_add(upfd, EVENT_READ);
-		if (!udp) event_add(conns[conn].downfd, EVENT_READ);
+		event_add(conns[conn].downfd, EVENT_READ);
 		servers[index].c++;
 		if (servers[index].status) {
 			servers[index].status = 0;
